
%{

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/ioctl.h>

#include "wan_ecmain.h"
#include <wanpipe_events.h>
#include "wanec_api.h"

extern wanec_client_t		ec_client;
extern wan_custom_conf_t	*custom_conf;
extern int			action;
extern char			yytext[];
extern char			**targv;
extern unsigned 		offset;
static int			start_channel = 0, range = 0;
extern int			gl_err;

extern int yylex(void);

unsigned long convert_str(char* str, int type);
static int is_channel(unsigned long);
void yyerror(char* msg);

static int wanec_client_param_name(char *key);
static int wanec_client_param_value(char *value);

extern int wanec_client_config(void);
extern int wanec_client_release(void);
extern int wanec_client_mode(int enable);
extern int wanec_client_bypass(int enable);
extern int wanec_client_opmode(int mode);
extern int wanec_client_modify(void);
extern int wanec_client_mute(int mode);
extern int wanec_client_dtmf(int enable);
extern int wanec_client_stats(int full);
extern int wanec_client_toneload(char *tone);
extern int wanec_client_toneunload(unsigned long tone_id);
extern int wanec_client_playout(int start, unsigned long tone_id);
extern int wanec_client_monitor(void);

%}

%union {
#define YYSTYPE YYSTYPE
	char*		str;
	unsigned long	val;
}


%token LOAD_TOKEN
%token UNLOAD_TOKEN
%token CONFIG_TOKEN
%token RELEASE_TOKEN
%token KILL_TOKEN
%token ENABLE_TOKEN
%token DISABLE_TOKEN
%token BYPASS_ENABLE_TOKEN
%token BYPASS_DISABLE_TOKEN
%token MODE_NORMAL_TOKEN
%token MODE_POWERDOWN_TOKEN
%token DTMF_ENABLE_TOKEN
%token DTMF_DISABLE_TOKEN
%token STATS_TOKEN
%token STATS_FULL_TOKEN
%token ALL_TOKEN
%token HELP_TOKEN
%token HELP1_TOKEN
%token MONITOR_TOKEN
%token MODIFY_TOKEN
%token TONE_LOAD_TOKEN
%token TONE_UNLOAD_TOKEN
%token PLAYOUT_START_TOKEN
%token PLAYOUT_STOP_TOKEN
%token DURATION_TOKEN
%token TEST_TOKEN
%token CUSTOM_PARAM_TOKEN
%token MUTE_TOKEN
%token UNMUTE_TOKEN
%token RIN_PORT_TOKEN
%token ROUT_PORT_TOKEN
%token SIN_PORT_TOKEN
%token SOUT_PORT_TOKEN
%token ALL_PORT_TOKEN

%token CHAR_STRING
%token DEC_CONSTANT
%token HEX_CONSTANT
%token DIAL_STRING

%%

start_args	: TEST_TOKEN
		  { action = WAN_EC_ACT_TEST; }
		| CHAR_STRING
		  { memcpy(ec_client.devname, $<str>1, strlen($<str>1));
		    wanec_api_init(); }
				command
		;

command		: CONFIG_TOKEN custom_param_list
		  { gl_err = wanec_client_config(); }
		| RELEASE_TOKEN
		  { gl_err = wanec_client_release(); }
		| KILL_TOKEN
		  { gl_err = wanec_client_release(); }
		| ENABLE_TOKEN		channel_map
		  { gl_err = wanec_client_mode(1); }
		| DISABLE_TOKEN		channel_map
		  { gl_err = wanec_client_mode(0); }
		| BYPASS_ENABLE_TOKEN	channel_map
		  { gl_err = wanec_client_bypass(1); }
		| BYPASS_DISABLE_TOKEN	channel_map
		  { gl_err = wanec_client_bypass(0); }
		| MODE_NORMAL_TOKEN	channel_map
		  { gl_err = wanec_client_opmode(1); }
		| MODE_POWERDOWN_TOKEN	channel_map
		  { gl_err = wanec_client_opmode(0); }
		| MODIFY_TOKEN		channel_map custom_param_list
		  { gl_err = wanec_client_modify(); }
		| MUTE_TOKEN		channel_map port_list
		  { gl_err = wanec_client_mute(1); }
		| UNMUTE_TOKEN		channel_map port_list
		  { gl_err = wanec_client_mute(2); }
		| DTMF_ENABLE_TOKEN	channel_map
		  { gl_err = wanec_client_dtmf(1); }
		| DTMF_DISABLE_TOKEN	channel_map
		  { gl_err = wanec_client_dtmf(0); }
		| STATS_TOKEN		stats_debug_args
		  { gl_err = wanec_client_stats(0); }
		| STATS_FULL_TOKEN	stats_debug_args
		  { gl_err = wanec_client_stats(1); }
		| TONE_LOAD_TOKEN	CHAR_STRING
		  { gl_err = wanec_client_toneload($<str>2); }
		| TONE_UNLOAD_TOKEN	DEC_CONSTANT
		  { gl_err = wanec_client_toneunload($<val>2); }
		| PLAYOUT_START_TOKEN	DEC_CONSTANT
		  { ec_client.channel = $<val>2; }
					DEC_CONSTANT
		  { gl_err = wanec_client_playout(1, $<val>4); }
						playout_args
		| PLAYOUT_STOP_TOKEN	DEC_CONSTANT
		  { ec_client.channel = $<val>2; }
					DEC_CONSTANT
		  { gl_err = wanec_client_playout(0, $<val>4); }
		| MONITOR_TOKEN		stats_debug_args
		  { gl_err = wanec_client_monitor(); }
		;

playout_args	: /* empty */
		| playout_args playout_arg
		;
		
playout_arg	: DURATION_TOKEN DEC_CONSTANT
		;
		
channel_map	: ALL_TOKEN
		  { ec_client.channel_map = 0xFFFFFFFF; }
		| channel_list
		;

channel_list	: /* empty */
		  { ec_client.channel_map = 0xFFFFFFFF; } 
		| DEC_CONSTANT
		  { is_channel($<val>1);
		    if (range){
			int	i=0;
			for(i=start_channel;i<=$<val>1;i++){
				ec_client.channel_map |= (1<<i);
			}
			start_channel=0;
		        range = 0;
		    }else{
		  	ec_client.channel_map |= (1 << $<val>1);
		    }
		  }
		| DEC_CONSTANT '-'
		  { is_channel($<val>1);
		    range = 1; start_channel = $<val>1; }
					channel_list
		| DEC_CONSTANT '.'
		  { is_channel($<val>1);
		    if (range){
			int	i=0;
			for(i=start_channel;i<=$<val>1;i++){
				ec_client.channel_map |= (1<<i);
			}
			start_channel=0;
		        range = 0;
		    }else{
		  	ec_client.channel_map |= (1 << $<val>1);
		    }
		  }
					channel_list
		;
		
stats_debug_args:
		  { ec_client.channel = 0; ec_client.channel_map = 0x00000000; }
		| DEC_CONSTANT
		  { is_channel($<val>1);
		    ec_client.channel = $<val>1; ec_client.channel_map = (1<<$<val>1); }
		;

port_list	: ALL_PORT_TOKEN
		  { ec_client.port_map = 
			WAN_EC_CHANNEL_PORT_SIN|WAN_EC_CHANNEL_PORT_SOUT|WAN_EC_CHANNEL_PORT_RIN|WAN_EC_CHANNEL_PORT_ROUT; }
		| ports
		;

ports		: port
		  { ec_client.port_map |= $<val>1; }
		| port
		  { ec_client.port_map |= $<val>1; }
			ports
		;
		
port		: SIN_PORT_TOKEN
		  { $<val>$ = WAN_EC_CHANNEL_PORT_SIN; }
		| SOUT_PORT_TOKEN
		  { $<val>$ = WAN_EC_CHANNEL_PORT_SOUT; }
		| RIN_PORT_TOKEN
		  { $<val>$ = WAN_EC_CHANNEL_PORT_RIN; }
		| ROUT_PORT_TOKEN
		  { $<val>$ = WAN_EC_CHANNEL_PORT_ROUT; }
		;
			
custom_param_list:
		| custom_param custom_param_list
		;

custom_param	: CUSTOM_PARAM_TOKEN CHAR_STRING '='
		  { wanec_client_param_name($<str>2); }
			CHAR_STRING
		  { wanec_client_param_value($<str>5); }
		;
%%

unsigned long convert_str(char* str, int type)
{
	unsigned long value = 0;
	switch(type){
	case DEC_CONSTANT:
		sscanf(str, "%lu", &value);
		break;
	case HEX_CONSTANT:
		sscanf(str, "%lx", &value);
		break;
	}
	return value;
}

static int is_channel(unsigned long channel)
{
	if (channel > 31){
		printf("ERROR: Channel number %ld is out of range !\n\n",
					channel);
		exit(1);
	}
	return 0;
}

void yyerror(char* msg)
{
	printf("> %s (argv=%s,offset=%d)\n", msg, *targv, offset);
	exit(1);
}


static int wanec_client_param_name(char *key)
{
	if (ec_client.conf.param_no == 0){
		ec_client.conf.params = malloc(sizeof(wan_custom_param_t));
		if (ec_client.conf.params == NULL){
			printf("ERROR: Failed to allocate structure for custom configuration!\n");
			return -EINVAL;
		}
		memset(ec_client.conf.params, 0, sizeof(wan_custom_param_t));
	}
	strlcpy(ec_client.conf.params[ec_client.conf.param_no].name, 
					key, MAX_PARAM_LEN); 
	return 0;
}

static int wanec_client_param_value(char *value)
{
	if (ec_client.conf.params == NULL){
		printf("ERROR: Custom configuration structure is not allocated!\n");
		return -EINVAL;
	} 
	strlcpy(ec_client.conf.params[ec_client.conf.param_no].sValue,
					value, MAX_VALUE_LEN); 
	ec_client.conf.param_no++;
	return 0;
}
