/******************************************************************************
** Copyright (c) 2005
**	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
**
** ============================================================================
** Sep 1, 2005		Alex Feldman	Initial version.
** Sep 2, 2005		Alex Feldman	Add option 'all' for channel 
**					selection.
******************************************************************************/

/******************************************************************************
**			   INCLUDE FILES
******************************************************************************/

#if !defined(__WINDOWS__)

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/queue.h>
#include <netinet/in.h>
#if defined(__LINUX__)
# include <linux/if.h>
# include <linux/types.h>
# include <linux/if_packet.h>
#endif

#else

#include "windows_config.h"

wanpipe_t	wanpipes[MAX_NUM_OF_WANPIPES];

char WAN_EC_DIR[MAX_PATH];
char WAN_EC_BUFFERS[MAX_PATH];
char windows_dir[MAX_PATH];

#endif//__WINDOWS__

#include "wan_ecmain.h"
#include <wanpipe_events.h>
#include "wanec_api.h"

/******************************************************************************
**			  DEFINES AND MACROS
******************************************************************************/

/******************************************************************************
**			STRUCTURES AND TYPEDEFS
******************************************************************************/

/******************************************************************************
**			   GLOBAL VARIABLES
******************************************************************************/
wanec_client_t	ec_client;

/******************************************************************************
** 			FUNCTION PROTOTYPES
******************************************************************************/
extern int wan_ec_args_parse_and_run(int, char**);
int wanec_client_config(void);
int wanec_client_release(void);
int wanec_client_mode(int enable);
int wanec_client_bypass(int enable);
int wanec_client_opmode(int mode);
int wanec_client_modify(void);
int wanec_client_mute(int mode);
int wanec_client_dtmf(int enable);
int wanec_client_stats(int full);
int wanec_client_toneload(char *tone);
int wanec_client_toneunload(unsigned long tone_id);
int wanec_client_playout(int start, unsigned long tone_id);
int wanec_client_monitor(void);

/******************************************************************************
** 			FUNCTION DEFINITIONS
******************************************************************************/

int main(int argc, char *argv[])
{
	int	err;
#if defined(__WINDOWS__)
	unsigned char	win_parse = 0;
	////////////////////////////////////////////////////////////////////////////
	//Get and parse command line arguments.
	////////////////////////////////////////////////////////////////////////////
	if(argc == 3 && !strcmp(argv[1], "-f")){
		//If "-f <file name> is provided, parse the file.
		win_parse = 1;
	}
	if(argc == 2 && !strcmp(argv[1], "-v")){
		//If "-v", print out version.
		win_parse = 1;
	}

	if(win_parse == 1){
		return parse_cmd_line_args(argc, argv);
	}
#endif

	memset(&ec_client, 0, sizeof(wanec_client_t));
	err = wan_ec_args_parse_and_run(argc, argv);
	if (!err){
		printf("\n\n");
	}
	if (ec_client.conf.param_no){
		free(ec_client.conf.params);
		ec_client.conf.params = NULL;
	}
	return err;
}

int wanec_client_config(void)
{
	wanec_api_config_t	conf;
	int 			err;

	memset(&conf, 0, sizeof(wanec_api_config_t));
	if (ec_client.conf.param_no){
		memcpy(&conf.conf, &ec_client.conf, sizeof(wan_custom_conf_t));
	}
	err = wanec_api_config(
			ec_client.devname,
			ec_client.verbose,
			&conf);
	return err;
}

int wanec_client_release(void)
{
	wanec_api_release_t	release;
	int			err;

	memset(&release, 0, sizeof(wanec_api_release_t));
	err = wanec_api_release(
			ec_client.devname,
			ec_client.verbose,
			&release);
	return err;
}

int wanec_client_mode(int enable)
{
	wanec_api_mode_t	mode;
	int			err;

	memset(&mode, 0, sizeof(wanec_api_mode_t));
	mode.enable	= enable;
	mode.channel_map= ec_client.channel_map;
	err = wanec_api_mode(
			ec_client.devname,
			ec_client.verbose,
			&mode);
	return err;
}

int wanec_client_bypass(int enable)
{
	wanec_api_bypass_t	bypass;
	int			err;

	memset(&bypass, 0, sizeof(wanec_api_bypass_t));
	bypass.enable		= enable;
	bypass.channel_map	= ec_client.channel_map;
	err = wanec_api_bypass(
			ec_client.devname,
			ec_client.verbose,
			&bypass);
	return err;
}

int wanec_client_opmode(int mode)
{
	wanec_api_opmode_t	opmode;
	int			err;

	memset(&opmode, 0, sizeof(wanec_api_opmode_t));
	opmode.mode		= mode;
	opmode.channel_map	= ec_client.channel_map;
	err = wanec_api_opmode(
			ec_client.devname,
			ec_client.verbose,
			&opmode);
	return err;
}


int wanec_client_modify()
{
	wanec_api_modify_t	modify;
	int			err;

	memset(&modify, 0, sizeof(wanec_api_modify_t));
	modify.channel_map	= ec_client.channel_map;
	if (ec_client.conf.param_no){
		memcpy(&modify.conf, &ec_client.conf, sizeof(wan_custom_conf_t));
	}
	err = wanec_api_modify(
			ec_client.devname,
			ec_client.verbose,
			&modify);
	return err;	
}

int wanec_client_mute(int mode)
{
	wanec_api_mute_t	mute;
	int			err;

	memset(&mute, 0, sizeof(wanec_api_mute_t));
	mute.mode		= mode;
	mute.channel_map	= ec_client.channel_map;
	mute.port_map		= ec_client.port_map;
	err = wanec_api_mute(
			ec_client.devname,
			ec_client.verbose,
			&mute);
	return err;
	
}

int wanec_client_dtmf(int enable)
{
	wanec_api_dtmf_t	dtmf;
	int			err;

	memset(&dtmf, 0, sizeof(wanec_api_dtmf_t));
	dtmf.enable		= enable;
	dtmf.channel_map	= ec_client.channel_map;
	dtmf.port_map		= WAN_EC_CHANNEL_PORT_SOUT;
	dtmf.type_map		= WAN_EC_TONE_PRESENT;
	err = wanec_api_dtmf(
			ec_client.devname,
			ec_client.verbose,
			&dtmf);
	return err;
}

int wanec_client_stats(int full)
{
	wanec_api_stats_t	stats;
	int			err;

	memset(&stats, 0, sizeof(wanec_api_stats_t));
	stats.full	= full;
	stats.channel	= ec_client.channel;
	stats.reset	= 0;	//reset
	err = wanec_api_stats(
			ec_client.devname,
			ec_client.verbose,
			&stats);
	return err;
}

int wanec_client_toneload(char *tone)
{
	wanec_api_toneload_t	toneload;
	int			err;

	memset(&toneload, 0, sizeof(wanec_api_toneload_t));
	toneload.ptone	= tone;
	err = wanec_api_tone_load(	
				ec_client.devname,
				ec_client.verbose,
				&toneload);
	if (!err){
		printf("Buffer tone index is %d!\n", toneload.tone_id);
	}
	return err;
}

int wanec_client_toneunload(unsigned long tone_id)
{
	wanec_api_toneunload_t	toneunload;
	int			err;

	memset(&toneunload, 0, sizeof(wanec_api_toneunload_t));
	toneunload.tone_id	= (unsigned int)tone_id;
	err = wanec_api_tone_unload(	
				ec_client.devname,
				ec_client.verbose,
				&toneunload);
	return err;
}

int wanec_client_playout(int start, unsigned long tone_id)
{
	wanec_api_playout_t	playout;
	int			err;

	memset(&playout, 0, sizeof(wanec_api_playout_t));
	playout.start		= start;
	playout.tone_id		= tone_id;
	playout.channel		= ec_client.channel;
	playout.duration	= 5000;
	err = wanec_api_playout(	
				ec_client.devname,
				ec_client.verbose,
				&playout);
	return err;
}

int wanec_client_monitor(void)
{
	wanec_api_monitor_t	monitor;
	int			err;

	memset(&monitor, 0, sizeof(wanec_api_monitor_t));
	monitor.channel		= ec_client.channel;
	err = wanec_api_monitor(	
				ec_client.devname,
				ec_client.verbose,
				&monitor);
	return err;
}
